// AnimatorDoc.cpp : implementation of the CAnimatorDoc class
//

#include "stdafx.h"
#include "Animator.h"

#include "AnimatorDoc.h"
#include "EditedField.h"
#include "ParameterDialog.h"
#include "BrowserControl.h"

#include "Dde.h"
#include "DDEML.h"

#include "vrmlparse.h"
#include "parse.tab.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CAnimatorDoc

IMPLEMENT_DYNCREATE(CAnimatorDoc, CDocument)

BEGIN_MESSAGE_MAP(CAnimatorDoc, CDocument)
	//{{AFX_MSG_MAP(CAnimatorDoc)
	ON_COMMAND(ID_FILE_OPEN, OnFileOpen)
	ON_COMMAND(ID_FILE_WRITEVRML, OnFileWriteVRML)
	ON_COMMAND(ID_FILE_PARAMS, OnFileParams)
	ON_COMMAND(ID_FILE_WRITEVRML_AS, OnFileWriteVRMLAs)
	ON_UPDATE_COMMAND_UI(ID_FILE_WRITEVRML, OnUpdateFileWriteVRML)
	ON_COMMAND(ID_UPDATE, OnUpdate)
	ON_COMMAND(ID_FILE_SAVE_AS, OnFileSaveAs)
	ON_COMMAND(ID_FILE_LOAD, OnFileLoad)
	ON_UPDATE_COMMAND_UI(ID_FILE_LOAD, OnUpdateFileLoad)
	ON_COMMAND(ID_FILE_BROWSER, OnFileBrowser)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CAnimatorDoc construction/destruction

CAnimatorDoc::CAnimatorDoc() :
	m_pScene(NULL), m_bLoop(FALSE), m_dLength(10.0), m_dStillPoint(-1.0),
	m_CurrentFilename(NULL), m_dwPidInst(0)
{
	m_bcDlg.m_sTempDir = "c:\\temp";
	m_FileVRML = m_bcDlg.m_sTempDir + "\\scene.wrl";
	m_FileTemplate = m_bcDlg.m_sTempDir + "\\scenetemplate.html";
}

CAnimatorDoc::~CAnimatorDoc()
{
}

BOOL CAnimatorDoc::OnNewDocument()
{
	if (!CDocument::OnNewDocument())
		return FALSE;

	// TODO: add reinitialization code here
	// (SDI documents will reuse this document)

	return TRUE;
}



/////////////////////////////////////////////////////////////////////////////
// CAnimatorDoc serialization

void CAnimatorDoc::Serialize(CArchive& ar)
{
	if (ar.IsStoring())
	{
		// TODO: add storing code here
	}
	else
	{
		// TODO: add loading code here
	}
}

/////////////////////////////////////////////////////////////////////////////
// CAnimatorDoc diagnostics

#ifdef _DEBUG
void CAnimatorDoc::AssertValid() const
{
	CDocument::AssertValid();
}

void CAnimatorDoc::Dump(CDumpContext& dc) const
{
	CDocument::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CAnimatorDoc commands

extern FILE* yyin;
extern node* root;

void add_names( nodelist* el, int* count );

void CAnimatorDoc::OnFileOpen() 
{
	CFileDialog dlg( TRUE, NULL, NULL, 0, "VRML Scenes (*.wrl)|*.wrl|Saved Animations (*.anm)|*.anm|All Files(*.*)|*.*||" );
	if ( dlg.DoModal() == IDOK ) {
		char buffer[80];
		FILE* vrmlfile = fopen( dlg.GetPathName(), "r" );
		fgets( buffer, 80, vrmlfile );

		yyin = vrmlfile;
		int ret = yyparse();
		fclose( vrmlfile );

		CEditedField::Reset();

		if ( ret ) 
		{
			TRACE( "VRML parse failed." );
			m_pScene = NULL;
		}
		else
		{
			int count = 1;
			m_pScene = scene;
			add_names( m_pScene, &count );
		}
	
		UpdateAllViews( NULL, UPDATE_NEW_VRML );
	}

}

void add_names( nodelist* el, int* count )
{
	field* f;
	node* e = el->node;
	char* temp;

	if ( el->use )
	{
		// nothing.
	}
	else
	{

		if ( e == NULL )
		{
			// nothing.
		}
		else if ( e->type == ROUTE )
		{
			// nothing.
		}
		else 
		{
			if ( e->name == NULL )
			{
				e->name = new char [ strlen(keyword_table[e->type])+10 ];
				strcpy( e->name, "_" );
				strcat( e->name, keyword_table[e->type] );
				sprintf( e->name + strlen(e->name), "%d", *count );
				++*count;
				for ( temp = e->name; *temp; ++temp )
					*temp = toupper( *temp );
			}
	  
			f = e->fields;
			while ( f )
			{
				f->parentnode = e;
				if ( f->nodetype == FT_NODE )
				{
					add_names( f->u.nl, count );
				}
				f = f->next;
			}
		}
	}

	if ( el->next )
		add_names( el->next, count );
}

void CAnimatorDoc::OnUpdateFileWriteVRML(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable( m_CurrentFilename != NULL );
}

void CAnimatorDoc::OnFileWriteVRML()
{
	if ( m_CurrentFilename )
		WriteVRML( m_CurrentFilename );
	else
		TRACE( "OnFileWriteVRML:  no current filename\n" );
}

void CAnimatorDoc::OnFileWriteVRMLAs() 
{
	if ( m_pScene == NULL ) return;

	CFileDialog dlg( FALSE, ".wrl", NULL, 0, "VRML Scenes (*.wrl)|*.wrl|All Files (*.*)|*.*||" );
	if ( dlg.DoModal() == IDOK ) {
		WriteVRML( (LPCSTR)dlg.GetPathName() );

		delete [] m_CurrentFilename;
		m_CurrentFilename = new char [ dlg.GetPathName().GetLength() + 1 ];
		strcpy( m_CurrentFilename, dlg.GetPathName() );
	}
}

void CAnimatorDoc::WriteVRML( const char *filename )
{
	FILE* vrmlfile = fopen( filename, "w" );

	fprintf( vrmlfile, "#VRML V2.0 utf8\n\n" );
 	reproduce( vrmlfile, m_pScene, 0, 0 );

	fprintf( vrmlfile, "\n#\n# begin generated section\n#\n\n" );
	fprintf( vrmlfile, "DEF _ANIMATION_TIMER TimeSensor {\n" );
	fprintf( vrmlfile, "  cycleInterval %f\n", m_dLength );
	fprintf( vrmlfile, "  loop %s\n", m_bLoop ? "FALSE" : "FALSE" );
	fprintf( vrmlfile, "}\n\n" );
		
	fprintf( vrmlfile, "DEF _TIME_DISTORTION ScalarInterpolator {\n" );
	fprintf( vrmlfile, "  key [ 0 1 ]\n" );
	fprintf( vrmlfile, "  keyValue [ 0 1 ]\n" );
	fprintf( vrmlfile, "}\n" );
	fprintf( vrmlfile, "ROUTE _ANIMATION_TIMER.fraction_changed TO _TIME_DISTORTION.set_fraction\n" );


	CEditedField::WriteAllVRML( vrmlfile );

	fprintf( vrmlfile, "\n##\n" );
	fprintf( vrmlfile, "## REMOVE THIS BLOCK TO DELETE THE CONTROL WIDGETS\n" );
	fprintf( vrmlfile, "##\n\n" );
	fprintf( vrmlfile, "DEF _X_CONTROL Transform { children [ Group { children [ DEF _START_BUTTON TouchSensor { }\n" );
	fprintf( vrmlfile, "Shape { geometry Sphere { radius 0.05 } appearance Appearance {	material Material {\n" );
	fprintf( vrmlfile, "diffuseColor 0 0.5 0 } } } ] } Group { children [ DEF _STOP_BUTTON TouchSensor { }\n" );
	fprintf( vrmlfile, "Transform { translation 0 0.1 0	children [ Shape { geometry Sphere { radius 0.05 }\n" );
	fprintf( vrmlfile, "appearance Appearance {	material Material { diffuseColor 0.5 0 0 } } } ] } ] } ] }\n" );

	fprintf( vrmlfile, "DEF _PROXIMITY ProximitySensor { size 2000 2000 2000 }\n" );
	fprintf( vrmlfile, "ROUTE _START_BUTTON.touchTime TO _ANIMATION_TIMER.startTime\n" );
	fprintf( vrmlfile, "ROUTE _STOP_BUTTON.touchTime TO _ANIMATION_TIMER.stopTime\n" );
	fprintf( vrmlfile, "\n##\n" );
	fprintf( vrmlfile, "## END CONTROL WIDGET BLOCK\n" );
	fprintf( vrmlfile, "##\n\n" );

	fclose ( vrmlfile );
}

void CAnimatorDoc::WriteTemplate( const char* filename, const char* vrmlfile,
								 float time )
{
	FILE* tmplfile = fopen( filename, "w" );
	
	fprintf( tmplfile, "<html><head><title>VRMLControl Template</title></head>\n" );

	fprintf( tmplfile, "<center>\n" );
	fprintf( tmplfile, "<embed src=\"%s\" border=0 width=80%% height=70%%><br>\n", vrmlfile );
	fprintf( tmplfile, "<applet code=\"VRMLControl\" name=\"VRMLControl\" width=80%% height=20%%>\n" );
	if ( time >= 0.0f )
	{
		fprintf( tmplfile, "<param name=\"startTime\" value=\"%f\">\n", time );
	}
	fprintf( tmplfile, "<param name=\"offsetX\" value=\"%f\">\n", m_bcDlg.m_fXOffset );
	fprintf( tmplfile, "<param name=\"offsetY\" value=\"%f\">\n", m_bcDlg.m_fYOffset );
	fprintf( tmplfile, "<param name=\"offsetZ\" value=\"%f\">\n", m_bcDlg.m_fZOffset );
	fprintf( tmplfile, "</applet></center></html>\n" );

	fclose( tmplfile );
}

void CAnimatorDoc::SendURL( const char* URL, BOOL newwindow )
{
	HCONV browser;
	HSZ appstring, topicstring;
	DWORD pidInst;

	pidInst = m_dwPidInst;

	appstring = DdeCreateStringHandle( pidInst, "iexplore", CP_WINANSI );
	topicstring = DdeCreateStringHandle( pidInst, "WWW_OpenURL", CP_WINANSI );
	browser = DdeConnect( pidInst, appstring, topicstring, NULL );
	if ( browser == 0 )
	{
		TRACE( "DdeConnect failed.\n" );
		switch ( DdeGetLastError( pidInst ) )
		{
		case DMLERR_DLL_NOT_INITIALIZED:
			TRACE( "DMLERR_DLL_NOT_INITIALIZED\n" );
			break;
		case DMLERR_INVALIDPARAMETER:
			TRACE( "DMLERR_INVALIDPARAMETER\n" );
			break;
		case DMLERR_NO_CONV_ESTABLISHED:
			TRACE( "DMLERR_NO_CONV_ESTABLISHED\n" );
			break;
		case DMLERR_NO_ERROR:
			TRACE( "DMLERR_NO_ERROR\n" );
			break;
		}

		AfxMessageBox( "Couldn't open connection to Internet Explorer." );
		return;
	}

	CString str;
	str.Format( "%s", URL );
	HSZ item = DdeCreateStringHandle( pidInst, str.GetBuffer(0), CP_WINANSI );
	TRACE( "sending [%s]\n", str );

	HDDEDATA ret2;
	ret2 = DdeClientTransaction( NULL, 0, browser, item, CP_WINANSI, 
		XTYP_REQUEST, 2000, NULL );
	TRACE( "transaction returned %d\n", ret2 );

	BOOL foo = DdeDisconnect( browser );
	TRACE( "disconnect returned %d\n", foo );
}


void CAnimatorDoc::UpdateBrowser( float time )
{	
	TRACE( "writing update at time %f to file %s\n", time, 
		m_FileTemplate );

	if ( time < 0.0f && m_dStillPoint >= 0.0 )
		time = (float)m_dStillPoint;
	WriteVRML( m_FileVRML );
	WriteTemplate( m_FileTemplate, m_FileVRML, time );

	if ( time >= 0.0f )
		m_dStillPoint = time;

	SendURL( m_FileTemplate );
}

void CAnimatorDoc::SetStillPoint( float time )
{
	m_dStillPoint = time;
}

void reproduce( FILE* of, nodelist* el, int depth, int newline )
{
	field* f;
	node* e = el->node;

	if ( el->use && e->type != ROUTE )
	{
		fprintf ( of, "USE %s\n", e->name );
	}
	else
	{

		if ( e == NULL )
		{
			fprintf ( of, "%*sNULL\n", newline ? depth*2 : 0, "" );
		}
		else if ( e->type == ROUTE )
		{
			fprintf ( of, "ROUTE %s TO %s\n", e->name, (char*)e->fields );
		}
		else 
		{
			if ( e->name )
				fprintf ( of, "%*sDEF %s %s {\n", newline ? depth*2 : 0, "",
					e->name, keyword_table[e->type] );
			else
				fprintf ( of, "%*s%s {\n", newline ? depth*2 : 0, "",
					keyword_table[e->type] );
	  
			f = e->fields;
			++depth;
			while ( f )
			{
				fprintf ( of, "%*s%s ", depth*2, "", keyword_table[f->key] );
				if ( f->nodetype == FT_NODE )
				{
					if ( f->u.nl && f->u.nl->next )
					{
						fprintf ( of, "[\n" );
						reproduce( of, f->u.nl, depth+1, 1 );
						fprintf ( of, "%*s]\n", depth*2, "" );
					}
					else
						reproduce( of, f->u.nl, depth, 0 );
				}
				else
				{
					fprintf ( of, "%s\n", f->u.text );
				}
				f = f->next;
			}
			--depth;

			if ( e )
				fprintf ( of, "%*s}\n", depth*2, "" );
		}
	}

	if ( el->next )
		reproduce( of, el->next, depth, 1 );
}
	       

void CAnimatorDoc::OnFileParams() 
{
	CParameterDialog dlg;
	
	dlg.m_dLength = m_dLength;
	dlg.m_bLoop = m_bLoop;
	if ( dlg.DoModal() == IDOK )
	{
		TRACE( "loop is now %d\n", dlg.m_bLoop );
		TRACE( "length is now %f\n", dlg.m_dLength );

		m_bLoop = dlg.m_bLoop;
		m_dLength = dlg.m_dLength;

		UpdateAllViews( NULL, UPDATE_PARAMETER_CHANGED );
	}
}



void CAnimatorDoc::OnUpdate() 
{
	UpdateBrowser();	
}

void CAnimatorDoc::OnFileSaveAs() 
{
	CFileDialog dlg( FALSE, ".anm", NULL, 0, "Animations (*.anm)|*.anm|All Files|*.*||" );

	if ( dlg.DoModal() == IDOK )
	{
		SaveAnimation( dlg.GetPathName() );
	}
}

void CAnimatorDoc::SaveAnimation( LPCTSTR filename )
{
	FILE* foo;

	foo = fopen( filename, "w" );
	if ( foo == NULL )
	{
		TRACE( "Couldn't write file for save.\n" );
		return;
	}

	fprintf( foo, "%d %f\n", m_bLoop, m_dLength );

	CEditedField::SaveAllCurves( foo );

	fclose ( foo );
}

void CAnimatorDoc::OnFileLoad() 
{
	CFileDialog dlg( TRUE, ".anm", NULL, 0, "Animations (*.anm)|*.anm|All Files|*.*||" );
	
	if ( dlg.DoModal() == IDOK )
	{
		LoadAnimation( dlg.GetPathName() );
		UpdateAllViews( NULL, UPDATE_CLOSE_ALL );
	}
}

void CAnimatorDoc::LoadAnimation( LPCTSTR filename )
{
	FILE *foo;
	int loop;

	foo = fopen( filename, "r" );
	if ( foo == NULL )
	{
		TRACE( "Couldn't open file for load.\n" );
		return;
	}

	fscanf( foo, "%d %lf\n", &loop, &m_dLength );
	m_bLoop = loop;

	CEditedField::LoadAllCurves( foo, this );

	fclose( foo );
}

void CAnimatorDoc::OnUpdateFileLoad(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable( m_pScene != NULL );
}

void CAnimatorDoc::OnFileBrowser() 
{
	if ( m_bcDlg.DoModal() == IDOK )
	{
		m_FileVRML = m_bcDlg.m_sTempDir + "\\scene.wrl";
		m_FileTemplate = m_bcDlg.m_sTempDir + "\\scenetemplate.html";
	}
}
