#include "stdafx.h"

#include "Animator.h"

#include "AnimatorDoc.h"
#include "PlotView.h"
#include "EditedField.h"
#include "Curve.h"
#include "LinearCurve.h"

CEditedField* CEditedField::listHead = NULL;
int CEditedField::editedFieldCount = 0;
int CEditedField::interpCounter = 0;

CEditedField::CEditedField( field* f, CAnimatorDoc* pDocument ) :
	pField(f), nDragPoint(-1), pDoc(pDocument)
{
	int i;

	next = listHead;
	prev = NULL;
	if ( listHead )	listHead->prev = this;

	listHead = this;
	++editedFieldCount;

	pCurveNames = NULL;
	pCurveColors = NULL;

	dyMin = 0.0;
	dyMax = 1.0;
	dxMin = 0.0;
	dxMax = pDoc->m_dLength;

	switch ( f->key)
	{
	case F_SHININESS:
	case F_AMBIENTINTENSITY:
	case F_TRANSPARENCY:
	case F_INTENSITY:
	case F_PRIORITY:
	case F_MINFRONT:
	case F_MAXFRONT:
	case F_MINBACK:
	case F_MAXBACK:
		nCurveCount = 1;
		break;

	case F_DIFFUSECOLOR:
	case F_SPECULARCOLOR:
	case F_EMISSIVECOLOR:
	case F_COLOR:
		nCurveCount = 3;
		pCurveNames = new char* [3];
		pCurveNames[0] = "red";
		pCurveNames[1] = "green";
		pCurveNames[2] = "blue";
		break;

	case F_SCALE:
		nCurveCount = 3;
		dyMin = 0.001;
		dyMax = 10.0;
		break;

	case F_TRANSLATION:
	case F_CENTER:
	case F_DIRECTION:
	case F_LOCATION:
		nCurveCount = 3;
		dyMin = -10.0;
		dyMax = 10.0;
		break;

	case F_ROTATION:
	case F_SCALEORIENTATION:
		dyMin = 0.0;
		dyMax = 2 * 3.14159;
		nCurveCount = 4;
		break;

	default:
		TRACE( "created CEditedField for bad field type!\n" );
		return;
	}
	nSelectedCurve = 0;

	if ( pCurveNames == NULL )
	{
		switch ( nCurveCount )
		{
		case 1:
			pCurveNames = new char* [1];
			pCurveNames[0] = keyword_table[f->key];
			break;

		case 3:
			pCurveNames = new char* [3];
			pCurveNames[0] = "x";
			pCurveNames[1] = "y";
			pCurveNames[2] = "z";
			break;

		case 4:
			pCurveNames = new char* [4];
			pCurveNames[0] = "x";
			pCurveNames[1] = "y";
			pCurveNames[2] = "z";
			pCurveNames[3] = "theta";
			break;
		}
	}

	if ( pCurveColors == NULL )
	{
		switch ( nCurveCount )
		{
		case 1:
			pCurveColors = new COLORREF [1];
			pCurveColors[0] = RGB(0,0,0);
			break;

		case 3:
			pCurveColors = new COLORREF [3];
			pCurveColors[0] = RGB(255,0,0);
			pCurveColors[1] = RGB(0,255,0);
			pCurveColors[2] = RGB(0,0,255);
			break;

		case 4:
			pCurveColors = new COLORREF [4];
			pCurveColors[0] = RGB(255,0,0);
			pCurveColors[1] = RGB(0,255,0);
			pCurveColors[2] = RGB(0,0,255);
			pCurveColors[3] = RGB(0,0,0);
			break;
		}
	}

	xControl = new double* [nCurveCount];
	yControl = new double* [nCurveCount];
	nControl = new int [nCurveCount];
	for ( i = 0; i < nCurveCount; ++i ) 
	{
		xControl[i] = new double [MAX_CONTROL_POINTS];
		yControl[i] = new double [MAX_CONTROL_POINTS];
		nControl[i] = 0;
	}

	pCurveCurves = new CCurve* [nCurveCount];
	for ( i = 0; i < nCurveCount; ++i )
		pCurveCurves[i] = CCurve::pCurves[0];

	f->editinfo = this;
}

CEditedField::~CEditedField()
{
	delete [] pCurveNames;
	delete [] pCurveColors;
	for ( int i = 0; i < nCurveCount; ++i )
	{
		delete [] xControl[i];
		delete [] yControl[i];
	}
	delete [] xControl;
	delete [] yControl;
	delete [] nControl;
	delete [] pCurveCurves;

	pField->editinfo = NULL;

	if ( next )	next->prev = prev;
	if ( prev ) prev->next = next;
	if ( listHead == this ) listHead = next;
	--editedFieldCount;
}

CEditedField* CEditedField::GetEditedField( field* f )
{
	CEditedField* p = listHead;

	while ( p )
	{
		if ( p->pField == f )
			return p;
		p = p->next;
	}
	return NULL;
}

void CEditedField::SaveAllCurves( FILE* foo )
{
	CEditedField* p = listHead;

	while( p )
	{
		p->SaveCurves( foo );
		p = p->next;
	}
}

void CEditedField::SaveCurves( FILE* foo )
{
	fprintf( foo, "*%s\n", pField->pathname );

	for ( int i = 0; i < nCurveCount; ++i )
	{
		fprintf( foo, "%s\n", pCurveNames[i] );
		fprintf( foo, "%s\n", pCurveCurves[i]->curvename );
		fprintf( foo, "%d", nControl[i] );
		for ( int j = 0; j < nControl[i]; ++j )
			fprintf( foo, " %.4f %.4f", xControl[i][j], yControl[i][j] );
		fprintf( foo, "\n" );
	}
	fprintf( foo, "-\n" );
}

void CEditedField::LoadAllCurves( FILE* foo, CAnimatorDoc* pDoc )
{
	char buffer[500];
	field *f;
	CEditedField* ef;
	CCurve** pc;
	double x, y;
	int t;

	CEditedField::Reset();

	while( fgets( buffer, 499, foo ) != NULL )
	{
		if ( buffer[0] != '*' ) continue;

		buffer[strlen(buffer)-1] = 0;
		TRACE( "searching for path [%s]\n", buffer+1 );

		f = FindPathname( buffer+1, pDoc->m_pScene );
		if ( f )
		{
			TRACE( "found it!  %08x\n", f );

			ef = new CEditedField( f, pDoc );

			while( fgets( buffer, 499, foo ) != NULL )
			{
				if ( buffer[0] == '-' ) break;
				buffer[strlen(buffer)-1] = 0;
				if ( buffer[0] == 0 ) continue;

				for ( int i = 0; i < ef->nCurveCount; ++i )
				{
					if ( strcmp( buffer, ef->pCurveNames[i] ) == 0 )
						break;
				}

				fgets( buffer, 499, foo );
				if ( i == ef->nCurveCount )
				{
					TRACE( "can't find curve %s\n", buffer );
					while ( fgets( buffer, 499, foo ) && buffer[strlen(buffer)-1] != '\n' );
					continue;
				}
				buffer[strlen(buffer)-1] = 0;

				pc = CCurve::pCurves;
				while ( *pc )
				{
					if ( strcmp( (*pc)->curvename, buffer ) == 0 )
						break;
					++pc;
				}
				if ( *pc == NULL )
				{
					TRACE( "can't find curve type %s\n", buffer );
					while ( fgets( buffer, 499, foo ) && buffer[strlen(buffer)-1] != '\n' );
					continue;
				}

				ef->pCurveCurves[i] = *pc;

				fscanf( foo, "%d", &t );
				ef->nControl[i] = 0;
				for ( int j = 0; j < t; ++j )
				{
					fscanf( foo, "%lf %lf\n", &x, &y );
					if ( x <= ef->dxMax )
					{
						ef->xControl[i][ef->nControl[i]] = x;
						ef->yControl[i][ef->nControl[i]] = y;
						++ef->nControl[i];
					}
				}
			}
		}
	}
}

field* CEditedField::FindPathname( char* target, nodelist* nl )
{
	node* e = nl->node;
	field *f;
	field *ret;

	if ( e && e->type != ROUTE )
	{
		f = e->fields;
		while( f )
		{
			if ( f->pathname && strcmp( f->pathname, target ) == 0 )
				return f;

			if ( f->realtype == MFNODE || f->nodetype == FT_NODE )
			{
				ret = FindPathname( target, f->u.nl );
				if ( ret ) return ret;
			}
			f = f->next;
		}
	}
	if ( nl->next )
	{
		ret = FindPathname( target, nl->next );
		if ( ret ) return ret;
	}
	return NULL;
}

BOOL CEditedField::IsFieldEditable( field* f )
{
	return ( f->key == F_SHININESS || f->key == F_AMBIENTINTENSITY || f->key == F_TRANSPARENCY || f->key == F_INTENSITY ||
			 f->key == F_PRIORITY || f->key == F_MINFRONT || f->key == F_MAXFRONT || f->key == F_MINBACK || 
			 f->key == F_MAXBACK ||
		     f->key == F_DIFFUSECOLOR || f->key == F_SPECULARCOLOR || f->key == F_EMISSIVECOLOR || f->key == F_COLOR ||
		     f->key == F_SCALE ||
			 f->key == F_TRANSLATION || f->key == F_CENTER || f->key == F_DIRECTION || f->key == F_LOCATION ||
			 f->key == F_DIRECTION ||
			 f->key == F_ROTATION || f->key == F_SCALEORIENTATION );
}

void CEditedField::Reset( void )
{
	listHead = NULL;
	editedFieldCount = 0;
	interpCounter = 0;
}

void CEditedField::SetLength( double newlength )
{
	CEditedField* p = listHead;

	while ( p )
	{
		for ( int i = 0; i < p->nCurveCount; ++i )
		{
			for ( int j = 0; j < p->nControl[i]; ++j )
			{
				if ( p->xControl[i][j] > newlength )
				{
					p->nControl[i] = j;
					break;
				}
			}
		}
		p->dxMax = newlength;
		p = p->next;
	}
}


void CEditedField::WriteAllVRML( FILE* vrmlfile )
{
	CEditedField* p = listHead;

	while ( p )
	{
		p->WriteVRML( vrmlfile );
		p = p->next;
	}
}

void CEditedField::WriteVRML( FILE* vrmlfile )
{
	int pos[4];
	CString key(""), keyValue("");
	CString temp;
	double x, y;
	int i;

	++interpCounter;
	switch ( pField->realtype )
	{
		case SFFLOAT:
			fprintf( vrmlfile, "DEF _I%d ScalarInterpolator {\n", interpCounter );
			break;

		case SFCOLOR:
			fprintf( vrmlfile, "DEF _I%d ColorInterpolator {\n", interpCounter );
			break;

		case SFVEC3F:
			fprintf( vrmlfile, "DEF _I%d PositionInterpolator {\n", interpCounter );
			break;

		case SFROTATION:
			fprintf( vrmlfile, "DEF _I%d OrientationInterpolator {\n", interpCounter );
			break;

		default:
			TRACE( "this should not happen.\n" );
			break;
	}

	CCurve* pTempCurves [4];
	for ( i = 0; i < nCurveCount; ++i )
	{
		pTempCurves[i] = pCurveCurves[i]->Duplicate();
		pTempCurves[i]->Generate( dxMin, dxMax, pDoc->m_bLoop, 
			nControl[i], xControl[i], yControl[i] );
		pos[i] = 0;
	}

	while ( pos[0] < pTempCurves[0]->n ||
		    (nCurveCount >= 2 && pos[1] < pTempCurves[1]->n) ||
		    (nCurveCount >= 3 && pos[2] < pTempCurves[2]->n) ||
		    (nCurveCount >= 4 && pos[3] < pTempCurves[3]->n) )
	{
		x = pTempCurves[0]->x[pos[0]];
		for ( i = 1; i < nCurveCount; ++i )
			if ( pTempCurves[i]->x[pos[i]] < x )
				x = pTempCurves[i]->x[pos[i]];

		//TRACE( "x is now %.3f\n", x );

		temp.Format( "%.3f", x / (dxMax-dxMin) );

		if ( key.GetLength() > 0 ) 
		{
			key += ", ";
			keyValue += ",";
		}
		key += temp;

		for ( i = 0; i < nCurveCount; ++i )
		{
			if ( pTempCurves[i]->x[pos[i]] == x )
			{
				y = pTempCurves[i]->y[pos[i]];
				//TRACE( "  curve %d:  exact\n", i );
			}
			else 
			{
				double x1, x2, y1, y2;
				x1 = pTempCurves[i]->x[pos[i]-1];
				y1 = pTempCurves[i]->y[pos[i]-1];
				x2 = pTempCurves[i]->x[pos[i]];
				y2 = pTempCurves[i]->y[pos[i]];

				if ( x1 == x2 )
				{
					//TRACE( "  curve %d:  interpolated duplicate\n", i );
					y = (y1+y2) / 2.0;
				}
				else
				{
					//TRACE( "  curve %d:  real interpolation\n", i );
					y = ( (x-x1) * y2 + (x2-x) * y1 ) / (x2-x1);
				}

			}

			//TRACE( "  curve %d y is   %.3f\n", i, y );

			temp.Format( " %.3f", y );
			keyValue += temp;

			while ( pTempCurves[i]->x[pos[i]] <= x && pos[i] < pTempCurves[i]->n ) 
			{
				++pos[i];
			}

		}
	}

	for ( i = 0; i < nCurveCount; ++i )
		delete pTempCurves[i];

	fprintf( vrmlfile, "  key [ %s ]\n  keyValue [ %s ]\n}\n",
		(LPCTSTR)key, (LPCTSTR)keyValue );

	fprintf ( vrmlfile, "ROUTE _TIME_DISTORTION.value_changed TO _I%d.set_fraction\n",
		interpCounter );
	fprintf ( vrmlfile, "ROUTE _I%d.value_changed TO %s.set_%s\n\n\n",
		interpCounter, pField->parentnode->name, keyword_table[pField->key] );
}

void CEditedField::Draw( CDC *pDC, int xoffset, int yoffset, int width, int height,
						double xstill )
{
	int i, j;

	nxMin = 60;
	nyMin = height-30;
	nxMax = width-10;
	nyMax = 40;

	CPen pen( PS_SOLID, 1, RGB(200,200,200) );
	pDC->SelectObject( pen );

	for ( i = 0; i <= 20; ++i )
	{
		j = nxMin + (nxMax-nxMin)*i/20;
		pDC->MoveTo( j + xoffset, nyMin + yoffset );
		pDC->LineTo( j + xoffset, nyMax + yoffset );
	}

	for ( i = 0; i <= 10; ++i )
	{
		j = nyMin + (nyMax-nyMin)*i/10;
		pDC->MoveTo( nxMin + xoffset, j + yoffset );
		pDC->LineTo( nxMax + xoffset, j + yoffset );
	}

	if ( xstill != -1.0 )
	{
		i = (int)( xstill * (double)(nxMax - nxMin) + nxMin );
		CPen thePen( PS_SOLID, 3, RGB(0,0,0) );
		pDC->SelectObject( thePen );
		pDC->MoveTo( i + xoffset, nyMin + yoffset );
		pDC->LineTo( i + xoffset, nyMax + yoffset );
	}

	CPoint center;
	CPoint offset( xoffset, yoffset );
	for ( i = 0; i < nCurveCount; ++i )
	{
		CPen thepen( PS_SOLID, 1, pCurveColors[i] );
		CBrush theBrush;
		theBrush.CreateSolidBrush( pCurveColors[i] );

		for ( j = 0; j < nControl[i]; ++j )
		{
			center = MapToPoint( xControl[i][j], yControl[i][j] ) + offset;

			pDC->SelectStockObject( NULL_PEN );
			pDC->SelectObject( theBrush );
			pDC->Ellipse( center.x - 3, center.y - 3, center.x + 4, center.y + 4 );

			pDC->SelectObject( thepen );
			pDC->SelectStockObject( HOLLOW_BRUSH );
			pDC->Ellipse( center.x - 5, center.y - 5, center.x + 5, center.y + 5 );
		}

		pDC->SelectObject( thepen );
		pDC->SelectStockObject( HOLLOW_BRUSH );
		pCurveCurves[i]->Generate( dxMin, dxMax, pDoc->m_bLoop, nControl[i], xControl[i], yControl[i] );

		if ( pCurveCurves[i]->n > 1 )
		{
			center = MapToPoint( pCurveCurves[i]->x[0], pCurveCurves[i]->y[0] ) + offset;
			pDC->MoveTo( center );

			for ( j = 1; j < pCurveCurves[i]->n; ++j )
				pDC->LineTo( MapToPoint( pCurveCurves[i]->x[j], pCurveCurves[i]->y[j] ) + offset );
		}
			
	}

	pDC->SelectStockObject( BLACK_PEN );
	pDC->SelectStockObject( HOLLOW_BRUSH );

	pDC->TextOut( xoffset+10, yoffset+10, 
		pField->pathname ? pField->pathname : "NULL field name" );

	CString str;
	pDC->SetTextAlign( TA_CENTER );
	str.Format( "%d", (int)dxMin );
	pDC->TextOut( nxMin, nyMin+8, str );
	str.Format( "%d", (int)dxMax );
	pDC->TextOut( nxMax, nyMin+8, str );
	pDC->SetTextAlign( TA_RIGHT );
	str.Format( "%.3f", dyMin );
	pDC->TextOut( nxMin-10, nyMin-5, str );
	str.Format( "%.3f", dyMax );
	pDC->TextOut( nxMin-10, nyMax-5, str );
	pDC->SetTextAlign( TA_LEFT );

	int legendstart = (int)(nyMin * 0.25 + nyMax * 0.75);
	int legendoffset = 15;
	for ( i = 0; i < nCurveCount; ++i )
	{
		pDC->SetTextColor( pCurveColors[i] );
		if ( nSelectedCurve == i )
		{
			CBrush theBrush( pCurveColors[i] );
			pDC->SelectObject( theBrush );
			pDC->SelectStockObject( NULL_PEN );
			pDC->Ellipse( 7, legendstart+4, 15, legendstart+12 );
			pDC->SelectStockObject( HOLLOW_BRUSH );
		}
		pDC->TextOut( 18, legendstart, pCurveNames[i] );
		legendstart += legendoffset;
	}

}

void CEditedField::ChooseCurve( int c )
{
	if ( c >= 0 && c < nCurveCount )
		nSelectedCurve = c;
}

void CEditedField::ChooseType ( int c )
{
	pCurveCurves[nSelectedCurve] = CCurve::pCurves[c];
}

double CEditedField::GetXFraction( double x )
{
	if ( x < nxMin ) return -1.0;
	else if ( x > nxMax ) return -1.0;
	else return (double)(x-nxMin)/(double)(nxMax-nxMin);
}

CPoint CEditedField::MapToPoint( double x, double y )
{
	CPoint p( (int)((x - dxMin) / (dxMax - dxMin) * (nxMax - nxMin) + nxMin),
		      (int)((y - dyMin) / (dyMax - dyMin) * (nyMax - nyMin) + nyMin) );
	return p;
}

void CEditedField::MapToCurve( CPoint point, double* x, double* y )
{
	*x = (double)(point.x - nxMin) / (nxMax - nxMin) * (dxMax - dxMin) + dxMin;
	*y = (double)(point.y - nyMin) / (nyMax - nyMin) * (dyMax - dyMin) + dyMin;
}

BOOL CEditedField::WithinGrid( const CPoint& point )
{
	return ( point.x >= nxMin && point.x <= nxMax &&
		     point.y <= nyMin && point.y >= nyMax );
}

void CEditedField::MakeWithinGrid( CPoint* point )
{
	if ( point->x < nxMin ) point->x = nxMin;
	else if ( point->x > nxMax ) point->x = nxMax;

	if ( point->y > nyMin ) point->y = nyMin;
	else if ( point->y < nyMax ) point->y = nyMax;
}

void CEditedField::MouseDown( CPoint point, UINT nFlags )
{
	if ( WithinGrid(point) )
	{
		int closest = -1, closestdist, dist;
		CPoint p;
		for ( int i = 0; i < nControl[nSelectedCurve]; ++i )
		{
			p = MapToPoint( xControl[nSelectedCurve][i], yControl[nSelectedCurve][i] );
			dist = (p.x-point.x)*(p.x-point.x) + (p.y-point.y)*(p.y-point.y);
			if ( dist < 100 && ( closest == -1 || dist < closestdist ) )
			{
				closest = i;
				closestdist = dist;
			}
		}

		double dx, dy;
		MapToCurve( point, &dx, &dy );

		if ( closest == -1 && ! ( nFlags & MK_CONTROL ) )
		{
			// add a new control point

			int i, j;
			for ( i = 0; i < nControl[nSelectedCurve] && 
				dx >= xControl[nSelectedCurve][i]; ++i );

			for ( j = nControl[nSelectedCurve]; j > i; --j )
			{
				xControl[nSelectedCurve][j] = xControl[nSelectedCurve][j-1];
				yControl[nSelectedCurve][j] = yControl[nSelectedCurve][j-1];
			}

			xControl[nSelectedCurve][i] = dx;
			yControl[nSelectedCurve][i] = dy;
			nDragPoint = i;
			++nControl[nSelectedCurve];
		}
		else if ( closest != -1 && (nFlags & MK_CONTROL) )
		{
			// delete a control point;

			for ( int j = closest; j < nControl[nSelectedCurve]-1; ++j )
			{
				xControl[nSelectedCurve][j] = xControl[nSelectedCurve][j+1];
				yControl[nSelectedCurve][j] = yControl[nSelectedCurve][j+1];
			}
			--nControl[nSelectedCurve];
			nDragPoint = -1;
		}
		else
		{
			// move an existing control point

			nDragPoint = closest;
		}
	}
}

void CEditedField::MouseDrag( CPoint point )
{
	if ( nDragPoint != -1 ) 
	{
		MakeWithinGrid( &point );
		double dx, dy;
		MapToCurve( point, &dx, &dy );

		while ( nDragPoint > 0 && xControl[nSelectedCurve][nDragPoint-1] > dx )
		{
			double temp;

			temp = xControl[nSelectedCurve][nDragPoint];
			xControl[nSelectedCurve][nDragPoint] = xControl[nSelectedCurve][nDragPoint-1];
			xControl[nSelectedCurve][nDragPoint-1] = temp;

			temp = yControl[nSelectedCurve][nDragPoint];
			yControl[nSelectedCurve][nDragPoint] = yControl[nSelectedCurve][nDragPoint-1];
			yControl[nSelectedCurve][nDragPoint-1] = temp;

			--nDragPoint;
		}

		while ( nDragPoint < nControl[nSelectedCurve]-1 &&
			xControl[nSelectedCurve][nDragPoint+1] < dx )
		{
			double temp;

			temp = xControl[nSelectedCurve][nDragPoint];
			xControl[nSelectedCurve][nDragPoint] = xControl[nSelectedCurve][nDragPoint+1];
			xControl[nSelectedCurve][nDragPoint+1] = temp;

			temp = yControl[nSelectedCurve][nDragPoint];
			yControl[nSelectedCurve][nDragPoint] = yControl[nSelectedCurve][nDragPoint+1];
			yControl[nSelectedCurve][nDragPoint+1] = temp;

			++nDragPoint;
		}

		xControl[nSelectedCurve][nDragPoint] = dx;
		yControl[nSelectedCurve][nDragPoint] = dy;

	}
}

void CEditedField::MouseUp( CPoint point )
{
	nDragPoint = -1;
}
