//
// copy and modify this file to generate a new curve type.

#include "stdafx.h"
#include "LinearCurve.h"

// constructor
CLinearCurve::CLinearCurve( void )
{
	// set "curvename" to a string naming this curve.  this
	// name is used in the curve type menu, and to identify
	// the curve type in saved animation files.

	// IT MUST NOT CONTAIN ANY NEWLINES.

	curvename = "Linear";
}

// Generate() is the main function for each curve.  It takes a set of 
// control points and generates a set of line segments approximating
// the desired curve.
//
// The output is a set of points placed in the instance variables x[] 
// and y[].  The number of points output is stored in the instance 
// variable n.  The Generate() function should first invoke delete on x
// and y to free any previously generated points, then allocate as
// many points as are necessary.
// 
// The first point must have an x-value equal to xmin.  The last point 
// must have an x-value equal to xmax.  The points must be sorted into
// nondecreasing order by x.
//
// If the wrap argument is true, then the first and last points should
// have equal y-values (although this is not enforced).  
//
// There are nc control points, given in the argument arrays xc[] and 
// yc[].
//
// If there are zero control points, you should return the two points 
// (xmin,0),(xmax,0).
//
// If there is one control point (x0,y0), you should return the two
// points (xmin,y0),(xmax,y0);

void CLinearCurve::Generate( double xmin, double xmax, BOOL wrap, 
							int nc, double* xc, double* yc )
{
	// first free any old points stored in this object.
	delete [] x;
	delete [] y;

	// now allocate room for new points.  for the linear curve type,
	// I never need more than nc+2 points in the output.
	x = new double [nc+2];
	y = new double [nc+2];

	// handle the case of one or two control points.  feel free
	// to just leave this code in each of your curves.
	if ( nc == 0 || nc == 1 )
	{
		x[0] = xmin; x[1] = xmax;
		y[1] = y[0] = (nc ? yc[0] : 0.0);
		n = 2;
		return;
	}

	x[0] = xmin;
	if ( wrap )
	{
		// if wrapping is on, I interpolate the y value at xmin and
		// xmax so that the slopes of the lines adjacent to the
		// wraparound are equal.

		y[0] = ( yc[0] * (xmax-xc[nc-1]) + yc[nc-1] * (xc[0]-xmin) ) /
			(xc[0] - xmin + xmax - xc[nc-1]);
	}
	else
	{
		// if wrapping is off, I make the first and last segments of
		// the curve horizontal.

		y[0] = yc[0];
	}

	// a linear curve is easy; the output is mostly just the control 
	// points.

	n = 1;
	for ( int i = 0; i < nc; ++i )
	{
		if ( xc[i] == x[n-1] ) continue;
		x[n] = xc[i];
		y[n] = yc[i];
		++n;
	}

	// set the endpoint based on the wrap flag.
	x[n] = xmax;
	if ( wrap )
		y[n] = y[0];
	else
		y[n] = y[n-1];
	++n;

	// n should contain the number of points you are outputting.
}

// remember to edit Curve.cpp to make sure your curve type
// is available!  (see the comments in that file.)