// PlotView.cpp : implementation file
//

#include "stdafx.h"
#include "Animator.h"
#include "AnimatorDoc.h"
#include "PlotView.h"
#include "EditedField.h"
#include "ScaleDialog.h"

#include "vrmlparse.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

int CPlotView::nEditorWidth = 600;
int CPlotView::nEditorHeight = 300;

/////////////////////////////////////////////////////////////////////////////
// CPlotView 

IMPLEMENT_DYNCREATE(CPlotView, CScrollView)

CPlotView::CPlotView() :
	nEditors(0), nTracking(-1), nMenuPane(-1), pTextFont(NULL)
{
}

CPlotView::~CPlotView()
{
}


BEGIN_MESSAGE_MAP(CPlotView, CScrollView)
	//{{AFX_MSG_MAP(CPlotView)
	ON_WM_SIZE()
	ON_WM_CONTEXTMENU()
	ON_WM_LBUTTONDOWN()
	ON_WM_LBUTTONUP()
	ON_WM_MOUSEMOVE()
	ON_COMMAND(IDM_EDIT_CLOSE, OnEditClose)
	ON_WM_ERASEBKGND()
	ON_COMMAND(IDM_EDIT_SCALE, OnEditScale)
	ON_COMMAND(IDM_EDIT_DELETE, OnEditDelete)
	//}}AFX_MSG_MAP
	ON_COMMAND_RANGE(IDM_EDIT_CURVEFIRST, IDM_EDIT_CURVELAST, OnEditCurve)
	ON_COMMAND_RANGE(IDM_EDIT_TYPEFIRST, IDM_EDIT_TYPELAST, OnEditType)
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CPlotView drawing

void CPlotView::OnInitialUpdate()
{
	CScrollView::OnInitialUpdate();

	CSize sizeTotal;
	// TODO: calculate the total size of this view
	sizeTotal.cx = nEditorWidth;
	sizeTotal.cy = nEditorHeight;
	SetScrollSizes( MM_TEXT, sizeTotal);
}

void CPlotView::OnDraw(CDC* pDC)
{
	CAnimatorDoc* pDoc = (CAnimatorDoc*) GetDocument();
	CString str;

	RECT clientRect;
	GetClientRect( &clientRect );
	CPoint topleft(0,clientRect.top), bottomleft(0,clientRect.bottom);
	pDC->DPtoLP( &topleft );
	pDC->DPtoLP( &bottomleft );

	CDC MDC;
	CBitmap b;
	VERIFY( MDC.CreateCompatibleDC( pDC ) );
	VERIFY( b.CreateCompatibleBitmap( pDC, nEditorWidth, nEditorHeight ) );
	MDC.SelectObject( b );
	MDC.SetMapMode( MM_TEXT );

	if ( nEditors == 0 )
	{
		pDC->FillSolidRect( &clientRect, RGB(255,255,255) );
		pDC->TextOut( 100, nEditorHeight/2, "no fields are being edited" );
	}
	else
	{
		for ( int i = 0; i < nEditors; ++i )
		{
			if ( nEditorHeight*i > bottomleft.y ||
				 nEditorHeight*(i+1) < topleft.y )
				 continue;

			InitializePaneDC( &MDC );
			pEditFields[i]->Draw( &MDC, 0, 0, nEditorWidth, nEditorHeight, pDoc->m_dStillPoint );

			pDC->BitBlt( 0, i*nEditorHeight, nEditorWidth, nEditorHeight, 
				&MDC, 0, 0, SRCCOPY );

		}

		if ( nEditorHeight*i < bottomleft.y )
			pDC->FillSolidRect( 0, nEditorHeight*i, 
				clientRect.right, bottomleft.y, RGB(255,255,255) );
	}

	MDC.SelectStockObject( SYSTEM_FONT );
	MDC.SelectStockObject( BLACK_PEN );
	MDC.SelectStockObject( NULL_BRUSH );
}

void CPlotView::InitializePaneDC( CDC* MDC )
{
	if ( pTextFont == NULL )
	{
		pTextFont = new CFont;
		pTextFont->CreateFont( 8, 0, 0, 0, 400, FALSE, FALSE, 0, ANSI_CHARSET, 
			OUT_DEFAULT_PRECIS,	CLIP_DEFAULT_PRECIS, DEFAULT_QUALITY, 
			DEFAULT_PITCH | FF_SWISS, "MS Sans Serif" );
	}

	CFont* pOldFont = (CFont*) MDC->SelectObject( pTextFont );

	MDC->FillSolidRect( 0, 0, nEditorWidth, nEditorHeight, RGB(255,255,255) );

	CPen pen( PS_SOLID, 3, RGB(200,200,200) );
	MDC->SelectObject( pen );
	MDC->MoveTo( 0, nEditorHeight-1 );
	MDC->LineTo( nEditorWidth, nEditorHeight-1 );
	MDC->MoveTo( 0, 0 );
	MDC->LineTo( nEditorWidth, 0 );

	MDC->SelectStockObject( BLACK_PEN );
}

/////////////////////////////////////////////////////////////////////////////
// CPlotView diagnostics

#ifdef _DEBUG
void CPlotView::AssertValid() const
{
	CScrollView::AssertValid();
}

void CPlotView::Dump(CDumpContext& dc) const
{
	CScrollView::Dump(dc);
}
#endif //_DEBUG

/////////////////////////////////////////////////////////////////////////////
// CPlotView message handlers

void CPlotView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
{

	switch ( lHint )
	{
	case UPDATE_NEW_VRML:
		nEditors = 0;
		nTracking = -1;
		nMenuPane = -1;
		Invalidate();
		break;

	case UPDATE_EDIT_FIELD:
		{
			EditField( (field*)pHint );
			CSize sizeTotal;
			sizeTotal.cx = nEditorWidth;
			sizeTotal.cy = nEditorHeight * CEditedField::editedFieldCount;
			SetScrollSizes( MM_TEXT, sizeTotal );
		}
		Invalidate();
		break;

	case UPDATE_PARAMETER_CHANGED:
		{
			CAnimatorDoc* pDoc = (CAnimatorDoc*) GetDocument();
			CEditedField::SetLength( pDoc->m_dLength );
		}
		Invalidate();
		break;

	case UPDATE_CLOSE_ALL:
		nEditors = 0;
		CSize sizeTotal;
		sizeTotal.cx = nEditorWidth;
		sizeTotal.cy = nEditorHeight * nEditors;
		SetScrollSizes( MM_TEXT, sizeTotal );

		Invalidate();
		break;
	}
	
}

void CPlotView::EditField( field* f )
{
	if ( nEditors >= MAX_EDITORS )
	{
		CString msg;
		msg.Format( "You have reached the maximum of %d edit panes.\n", MAX_EDITORS );
		MessageBox( msg, "Error" );
		return;
	}

	CEditedField* pEdit = CEditedField::GetEditedField( f );
	if ( pEdit ) 
	{
		for ( int i = 0; i < nEditors; ++i )
			if ( pEdit == pEditFields[i] )
			{
				MessageBox( "There already is an edit pane for that field.\n" );
				return;
			}
	}
	else
	{

		if ( CEditedField::IsFieldEditable( f ) == FALSE )
		{
			MessageBox( "That field type is not editable.\n" );
			return;
		}

		pEdit = new CEditedField( f, (CAnimatorDoc*)GetDocument() );

		CAnimatorDoc* pDoc = (CAnimatorDoc*) GetDocument();
	}

	pEditFields[nEditors++] = pEdit;

}


void CPlotView::OnSize(UINT nType, int cx, int cy) 
{
	CScrollView::OnSize(nType, cx, cy);
	
	if ( cx > 0 )
	{
		nEditorWidth = cx;
		CSize sizeTotal;
		sizeTotal.cx = nEditorWidth;
		sizeTotal.cy = nEditorHeight * nEditors;
		SetScrollSizes( MM_TEXT, sizeTotal );
	}
}

void CPlotView::OnContextMenu(CWnd* pWnd, CPoint point) 
{
	CPoint local;
	CRect me;
	int i, selected;

	GetWindowRect( &me );
	//TRACE( "windowrect is %d %d %d %d\n", me.left, me.top, me.right, me.bottom );
	//TRACE( "screen point is %d %d\n", point.x, point.y );
	local.x = point.x - me.left;
	local.y = point.y - me.top;
	//TRACE( "device point at %d,%d\n", local.x, local.y );

	CClientDC dc( this );
	OnPrepareDC( &dc );
	dc.DPtoLP( &local );

	//TRACE( "logical point at %d,%d\n", local.x, local.y );

	int pane = local.y / nEditorHeight;
	if ( pane >= nEditors ) return;

	nMenuPane = pane;
	CEditedField* ef = pEditFields[pane];

	CMenu menu;
	menu.LoadMenu( IDR_EDIT_POPUP );
	CMenu* popup = menu.GetSubMenu( 0 );

	for ( i = 0; i < ef->nCurveCount; ++i )
	{
		char* itemstring = new char [6+strlen(ef->pCurveNames[i])];
		strcpy( itemstring, "Edit " );
		strcat( itemstring, ef->pCurveNames[i] );
		popup->AppendMenu( MF_STRING, IDM_EDIT_CURVEFIRST+i, itemstring );
	}
	popup->CheckMenuRadioItem( IDM_EDIT_CURVEFIRST, IDM_EDIT_CURVELAST, 
		IDM_EDIT_CURVEFIRST + ef->nSelectedCurve, MF_BYCOMMAND );

	popup->AppendMenu( 0, MF_SEPARATOR );

	selected = -1;
	for ( i = 0; CCurve::pCurves[i]; ++i )
	{
		popup->AppendMenu( MF_STRING, IDM_EDIT_TYPEFIRST+i, CCurve::pCurves[i]->curvename );
		if ( ef->pCurveCurves[ef->nSelectedCurve] == CCurve::pCurves[i] )
			selected = i;
	}
	if ( selected != -1 )
		popup->CheckMenuRadioItem( IDM_EDIT_TYPEFIRST, IDM_EDIT_TYPELAST, 
			IDM_EDIT_TYPEFIRST+selected, MF_BYCOMMAND );
	else
		MessageBox( "no curve type is selected:  this should not have happened.\n" );

	popup->TrackPopupMenu( TPM_LEFTALIGN | TPM_RIGHTBUTTON, point.x, point.y, this );
}


void CPlotView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	CAnimatorDoc* pDoc = (CAnimatorDoc*)GetDocument();
	
	CClientDC dc( this );
	OnPrepareDC( &dc );
	dc.DPtoLP( &point );

	int pane = point.y / nEditorHeight;
	if ( pane >= nEditors ) return;

	if ( nFlags & MK_SHIFT )
	{
		pDoc->SetStillPoint( (float)(pEditFields[pane]->GetXFraction( point.x )) );
		Invalidate();
		return;
	}

	if ( nTracking == -1 )
	{
		nTracking = pane;
		pEditFields[pane]->MouseDown( CPoint(point.x, point.y - pane * nEditorHeight), nFlags );

		CDC MDC;
		CBitmap b;
		VERIFY( MDC.CreateCompatibleDC( &dc ) );
		VERIFY( b.CreateCompatibleBitmap( &dc, nEditorWidth, nEditorHeight ) );
		MDC.SelectObject( b );
		MDC.SetMapMode( MM_TEXT );

		InitializePaneDC( &MDC );

		pEditFields[pane]->Draw( &MDC, 0, 0, nEditorWidth, nEditorHeight, pDoc->m_dStillPoint );
		dc.BitBlt( 0, nTracking * nEditorHeight, nEditorWidth, nEditorHeight,
				&MDC, 0, 0, SRCCOPY );

		MDC.SelectStockObject( SYSTEM_FONT );

		SetCapture();

	}
}


void CPlotView::OnLButtonUp(UINT nFlags, CPoint point) 
{
	if ( nTracking != -1 )
	{
		CAnimatorDoc* pDoc = (CAnimatorDoc*)GetDocument();

		CClientDC dc( this );
		OnPrepareDC( &dc );
		dc.DPtoLP( &point );

		CPoint panePoint( point.x, point.y - nTracking * nEditorHeight );
		pEditFields[nTracking]->MouseDrag( panePoint );
		pEditFields[nTracking]->MouseUp( panePoint );

		CDC MDC;
		CBitmap b;
		VERIFY( MDC.CreateCompatibleDC( &dc ) );
		VERIFY( b.CreateCompatibleBitmap( &dc, nEditorWidth, nEditorHeight ) );
		MDC.SelectObject( b );
		MDC.SetMapMode( MM_TEXT );

		InitializePaneDC( &MDC );

		pEditFields[nTracking]->Draw( &MDC, 0, 0, nEditorWidth, nEditorHeight, pDoc->m_dStillPoint );
		dc.BitBlt( 0, nTracking * nEditorHeight, nEditorWidth, nEditorHeight,
				&MDC, 0, 0, SRCCOPY );

		MDC.SelectStockObject( SYSTEM_FONT );

		nTracking = -1;
		ReleaseCapture();
	}

	//CScrollView::OnLButtonUp(nFlags, point);
}

void CPlotView::OnMouseMove(UINT nFlags, CPoint point) 
{
	if ( nTracking != -1 )
	{
		CAnimatorDoc* pDoc = (CAnimatorDoc*) GetDocument();

		CClientDC dc( this );
		OnPrepareDC( &dc );
		dc.DPtoLP( &point );

		pEditFields[nTracking]->MouseDrag( CPoint(point.x, 
			point.y - nTracking * nEditorHeight) );

		CDC MDC;
		CBitmap b;
		VERIFY( MDC.CreateCompatibleDC( &dc ) );
		VERIFY( b.CreateCompatibleBitmap( &dc, nEditorWidth, nEditorHeight ) );
		MDC.SelectObject( b );
		MDC.SetMapMode( MM_TEXT );

		InitializePaneDC( &MDC );

		pEditFields[nTracking]->Draw( &MDC, 0, 0, nEditorWidth, nEditorHeight, pDoc->m_dStillPoint );
		dc.BitBlt( 0, nTracking * nEditorHeight, nEditorWidth, nEditorHeight,
				&MDC, 0, 0, SRCCOPY );

		MDC.SelectStockObject( SYSTEM_FONT );
	}
}

void CPlotView::OnEditClose() 
{
	for ( int i = nMenuPane; i < nEditors-1; ++i )
		pEditFields[i] = pEditFields[i+1];
	--nEditors;

	CSize sizeTotal;
	sizeTotal.cx = nEditorWidth;
	sizeTotal.cy = nEditorHeight * nEditors;
	SetScrollSizes( MM_TEXT, sizeTotal );

	Invalidate();
}

void CPlotView::OnEditCurve( UINT nID )
{
	pEditFields[nMenuPane]->ChooseCurve( nID - IDM_EDIT_CURVEFIRST );
	Invalidate();
}

void CPlotView::OnEditType( UINT nID )
{
	pEditFields[nMenuPane]->ChooseType ( nID - IDM_EDIT_TYPEFIRST );
	Invalidate();
}

BOOL CPlotView::OnCommand(WPARAM wParam, LPARAM lParam) 
{
	TRACE( "wParam: %x\n", wParam );
	
	return CScrollView::OnCommand(wParam, lParam);
}

BOOL CPlotView::OnEraseBkgnd(CDC* pDC) 
{
	return TRUE;  
	//return CScrollView::OnEraseBkgnd(pDC);
}

void CPlotView::OnEditScale() 
{
	CScaleDialog dlg;

	dlg.m_dMinimum = pEditFields[nMenuPane]->dyMin;
	dlg.m_dMaximum = pEditFields[nMenuPane]->dyMax;
	if ( dlg.DoModal() == IDOK )
	{
		pEditFields[nMenuPane]->dyMin = dlg.m_dMinimum;
		pEditFields[nMenuPane]->dyMax = dlg.m_dMaximum;

		Invalidate();
	}
}

void CPlotView::OnEditDelete() 
{
	delete pEditFields[nMenuPane];
	for ( int i = nMenuPane; i < nEditors-1; ++i )
		pEditFields[i] = pEditFields[i+1];
	--nEditors;
	Invalidate();
}
