#include "stdafx.h"
#include "QuadraticSubdivisionCurve.h"

CQuadraticSubdivisionCurve::CQuadraticSubdivisionCurve( void )
{
	curvename = "Quadratic subdivision";
}

void CQuadraticSubdivisionCurve::Generate( double xmin, double xmax, BOOL wrap,
										   int nc, double* xc, double* yc )
{
	int i, j, k;
	int p;
	double tempx1, tempy1, tempx2, tempy2;

	delete [] x;
	delete [] y;

	if ( nc == 0 || nc == 1 )
	{
		x = new double [2];
		y = new double [2];
		x[0] = xmin;
		x[1] = xmax;
		y[0] = y[1] = (nc ? yc[0] : 0.0);
		n = 2;
		return;
	}
	
	if ( wrap )
	{
		double* sx = new double [nc*8];
		double* sy = new double [nc*8];
		x = new double [nc*8+2];
		y = new double [nc*8+2];
		double* tx = x + 1;
		double* ty = y + 1;

		memcpy( sx, xc, nc * sizeof ( double ) );
		memcpy( sy, yc, nc * sizeof ( double ) );
		p = nc;

		// subdivide

		for ( j = 0; j < 3; ++j )
		{
			tempx1 = 0.75 * (sx[p-1]-(xmax-xmin)) + 0.25 * sx[0];
			tempy1 = 0.75 * sy[p-1] + 0.25 * sy[0];

			tempx2 = 0.25 * (sx[p-1]-(xmax-xmin)) + 0.75 * sx[0];
			tempy2 = 0.25 * sy[p-1] + 0.75 * sy[0];

			if ( tempx1 >= xmin )
			{
				tx[0] = tempx1;
				ty[0] = tempy1;
				tx[1] = tempx2;
				ty[1] = tempy2;
				k = 2;
			}
			else if ( tempx2 >= xmin )
			{
				tx[p*2-1] = tempx1 + (xmax-xmin);
				ty[p*2-1] = tempy1;
				tx[0] = tempx2;
				ty[0] = tempy2;
				k = 1;
			}
			else
			{
				tx[p*2-2] = tempx1 + (xmax-xmin);
				ty[p*2-2] = tempy1;
				tx[p*2-1] = tempx2 + (xmax-xmin);
				ty[p*2-1] = tempy2;
				k = 0;
			}					

			for ( i = 0; i < p-1; ++i )
			{
				tx[k] = 0.75 * sx[i] + 0.25 * sx[i+1];
				ty[k] = 0.75 * sy[i] + 0.25 * sy[i+1];
				++k;

				tx[k] = 0.25 * sx[i] + 0.75 * sx[i+1];
				ty[k] = 0.25 * sy[i] + 0.75 * sy[i+1];
				++k;
			}

			p *= 2;
			memcpy( sx, tx, p * sizeof ( double ) );
			memcpy( sy, ty, p * sizeof ( double ) );
		}

		delete [] sx;
		delete [] sy;

		x[0] = xmin;
		x[k+1] = xmax;

		y[0] = ( y[1] * (xmax-x[k]) + y[k] * (x[1]-xmin) ) / ( xmax-x[k]+x[1]-xmin );
		y[k+1] = y[0];

		n = k + 2;
	}
	else
	{
		double* sx = new double [(nc+2)*8];
		double* sy = new double [(nc+2)*8];
		double* tx = new double [(nc+2)*8];
		double* ty = new double [(nc+2)*8];

		sx[0] = xmin;
		sy[0] = yc[0];
		for ( i = 0; i < nc; ++i )
		{
			sx[i+1] = xc[i];
			sy[i+1] = yc[i];
		}
		sx[nc+1] = xmax;
		sy[nc+1] = yc[nc-1];
		p = nc + 2;

		// subdivide

		for ( j = 0; j < 3; ++j )
		{
			tx[0] = sx[0]; 
			ty[0] = sy[0];

			k = 1;
			for ( i = 0; i < p-1; ++i )
			{
				tx[k] = 0.75 * sx[i] + 0.25 * sx[i+1];
				ty[k] = 0.75 * sy[i] + 0.25 * sy[i+1];
				++k;

				tx[k] = 0.25 * sx[i] + 0.75 * sx[i+1];
				ty[k] = 0.25 * sy[i] + 0.75 * sy[i+1];
				++k;
			}
			tx[k] = sx[p-1];
			ty[k] = sy[p-1];
			++k;

#if 0
			TRACE( "one iteration: produced %d points\n", k );
#endif

			memcpy( sx, tx, k * sizeof ( double ) );
			memcpy( sy, ty, k * sizeof ( double ) );
			p = k;
		}

		delete [] tx;
		delete [] ty;

		x = sx;
		y = sy;
		n = p;
	}
}