#include "stdafx.h"

#include <stdio.h>

#include "vrmlparse.h"
#include "parse.tab.h"

named_node* directory = NULL;

#ifndef WIN32
main( int argc, char **argv )
{
     char buffer[80];

     initialize_keyword_table();
     
     if ( argc == 2 )
	  yyin = fopen( argv[1], "r" );
     else
	  yyin = stdin;

     if ( yyin == NULL )
     {
	  fprintf ( stderr, "couldn't open input file.\n" );
	  exit(1);
     }
     
     fgets( buffer, 80, yyin );

     /*yydebug = 1;*/
     yyparse();

     printf ( "----- traversal\n" );
     traverse( scene, 0 );

     printf ( "----- reproduction\n" );
     reproduce( scene, 0, 0 );
}
#endif

int yyerror( char *s )
{
     fprintf( stderr, "yyerror: %s\n", s );
     return 1;
}

int yywrap( void )
{
     return 1;
}

void traverse( nodelist* el, int depth )
{
     field* f;
     node* e = el->node;
     
     printf ( "%*s%s %x\n", depth*4, "", e ? keyword_table[e->type] : "NULL", e );

     if ( e && e->type != ROUTE )
     {
	  /** traverse field list **/
	  
	  f = e->fields;
	  while ( f )
	  {
	       printf ( "%*s%s\n", depth*4+2, "", keyword_table[f->key] );
	       if ( f->nodetype == FT_NODE )
		    traverse( f->u.nl, depth+1 );
	       f = f->next;
	  }
     }

     if ( el->next )
	  traverse( el->next, depth );
}
	       
nodelist* reverse_nodelist ( nodelist* head )
{
     nodelist * newhead, * oldhead, * temp;

     if ( head == NULL ) return NULL;
     
     newhead = head;
     oldhead = head->next;
     head->next = NULL;

     while( oldhead )
     {
          temp = oldhead->next;
          oldhead->next = newhead;
          newhead = oldhead;
          oldhead = temp;
     }

     return newhead;
}

field* reverse_field ( field* head )
{
     field* newhead, * oldhead, * temp;

     if ( head == NULL ) return NULL;
     
     newhead = head;
     oldhead = head->next;
     head->next = NULL;

     while( oldhead )
     {
          temp = oldhead->next;
          oldhead->next = newhead;
          newhead = oldhead;
          oldhead = temp;
     }

     return newhead;
}

void def_name( char* name, node* object )
{
     named_node* p = (named_node *) new named_node;
     p->name = name;
     p->object = object;
     p->next = directory;
     directory = p;
}

node* use_name( char* name )
{
     named_node* p = directory;
     while( p )
     {
          if ( strcmp( name, p->name ) == 0 )
               return p->object;
          p = p->next;
     }
     return NULL;
}
