#ifndef _VRMLPARSE_H
#define _VRMLPARSE_H

#include <stdio.h>

typedef struct _node
{
     int type;
     char* name;
     struct _field* fields;
} node;

typedef struct _nodelist
{
     struct _nodelist* next;
     int use;
     struct _node* node;
} nodelist;

typedef struct _field
{
     int key;
     int nodetype;
     union
     {
	  char* text;
	  nodelist* nl;
     } u;
     int realtype;
	 node* parentnode;
#ifdef WIN32
	 char* pathname;
     void* editinfo;
#endif
     struct _field* next;
} field;

#define FT_NODE  1
#define FT_TEXT  2

typedef struct _named_node
{
     char* name;
     node* object;
     struct _named_node* next;
} named_node;

extern FILE* yyin;
extern nodelist* scene;
extern int yydebug;

int yyparse( void );
int yyerror( char * );
int yywrap( void );
int yylex( void );

void initialize_keyword_table( void );
extern char* keyword_table[];

nodelist* reverse_nodelist( nodelist* head );
field* reverse_field( field* head );
void traverse( nodelist* e, int depth );
void reproduce( FILE*, nodelist* e, int depth, int newline );
void def_name( char* name, node* object );
node* use_name( char* name );

#ifdef WIN32
#include <malloc.h>
#define alloca _alloca
#define YY_NEVER_INTERACTIVE 1
#endif

#endif
